/*
  Algorithms IE171- Lab11/Cryptography
  Author:  Menal Guzelsoy - 04/13 Spring 2004
*/

/*------------------------------------------------------------------------*/

#include "rsa.hpp"
#include <cmath>

/**************************************************************************/

RSA::RSA(): 
   N_(0), 
   encBlockSize_(4), decBlockSize_(5)
{
   initialize_();
}

/**************************************************************************/

RSA::RSA(ifstream &keyFile): 
   N_(0), 
   encBlockSize_(4), decBlockSize_(5)
{
   string temp;

   if(keyFile){
      keyFile>>temp>>publicKey_.first>>publicKey_.second;
      keyFile>>temp>>privateKey_.first>>privateKey_.second;
   }
   else{
      cout<<"RSA(): Unable to open the key file!\n"<<endl;
      exit(1);
   }

   initialize_();
}

/**************************************************************************/

void RSA::generateKeys()
{

   INT phi;   
   /* create two random primes */
   pair<INT, INT> factors = generatePrimeNumbers_();

   /* create the public key */
   
   publicKey_.first = getRandNumPrimeToPhi_(factors);
   publicKey_.second = factors.first * factors.second;
   
   /* now create the private key */

   phi = (factors.first - 1)*(factors.second - 1);
   privateKey_.first = getInverse_(publicKey_.first, phi);
   privateKey_.second = publicKey_.second;

   /* publish only the public key! */

   cout<<"Public Key = ("<<publicKey_.first<<", "<<publicKey_.second<<")"
       <<endl;
   
   /* keep the keys in a separate file for private use */

   ofstream os("myKeys.txt");
   os<<"Public_Key: "<<endl<<publicKey_.first<<" "<<publicKey_.second<<endl;
   os<<"Private_Key: "<<endl<<privateKey_.first<<" "<<privateKey_.second
     <<endl;
   os.close();
}

/**************************************************************************/

string RSA::encode(string textToCode)
{
   return encode(textToCode, publicKey_); 
}

/**************************************************************************/

string RSA::encode(string textToCode, pair<INT, INT> pubKey, int encSize, 
		   int decSize)
{
   int i, k, l, temp, length;
   string cipherText;
   INT P, C, e = pubKey.first, n = pubKey.second;

   if(!pubKey.first || !pubKey.second){
      cout<<"encode(): Public key entitites can not have null values!"<<endl;
      exit(1);
   }

   /* If the block sizes are not defined, use the default ones */
   k = encSize?encSize:encBlockSize_;
   l = decSize?decSize:decBlockSize_;

   /* Divide the whole string into k-size blocks.If the length of the string 
      is not a multiple of k, then add space characters to the end as many 
      as required */

   length = textToCode.length();
   temp = length % k;
   if(temp){
      temp = k - temp;
      for(i = 0; i < temp; i++) {
	 textToCode += " ";
      }
   }

   temp = (length + temp);

   cout<<"Encoding..."<<endl;

   for(i = 0; i < temp ; i = i + k){

      P = convertToInt_(textToCode.substr(i, k));
   
      C = getPower_(P, e, n);

      cipherText += convertToStr_(C, l);

      cout<<"Block "<<(i/k)+1<<" encoded..."<<endl;      
   }

  return cipherText;
}
      
/**************************************************************************/

string RSA::decode(string textToCode)
{
   return decode_(textToCode, privateKey_);
}

/**************************************************************************/

string RSA::breakCode(string textToCode, pair<INT, INT> pubKey, int encSize, 
		      int decSize)
{
   INT phi;
   pair<INT, INT> priKey;
   pair<INT, INT> factors;

   factors = factorize_(pubKey.second);
   phi = (factors.first - 1)*(factors.second - 1);

   priKey.first = getInverse_(pubKey.first, phi);
   priKey.second = pubKey.second;
   
   return decode_(textToCode, priKey, encSize, decSize);
}
   
/**************************************************************************/

string RSA::decode_(string textToCode, pair<INT, INT> priKey, int encSize, 
		    int decSize)
{
   INT num, P, C;
   string plainText;
   int i, length, k, l;
   
   if(!priKey.first || !priKey.second){
      cout<<"decode_(): Private Key entitites can not have null values!"<<endl;
      exit(1);
   }
   
   k = encSize?encSize:encBlockSize_;
   l = decSize?decSize:decBlockSize_;

   length = textToCode.length();
   
   if(length%l) {
      cout<<"decode(): length of the decoded text is expected to be a "
	  <<"multiple of decoding block size "<<l<<endl;
      exit(1);
   }
   
   cout<<"Decoding..."<<endl;
   for(i = 0; i < length; i = i + l){
      
      C = convertToInt_(textToCode.substr(i, l));
      P = getPower_(C, priKey.first, priKey.second);
      plainText += convertToStr_(P, k);

      cout<<"Block "<<(i/l)+1<<" decoded..."<<endl;

   }
   return plainText;
}   

/**************************************************************************/

void RSA::initialize_()
{

   int temp;
   int i(0), j(0);

   /* First map the following useful characters including the numbers */

   char * characters = " !?'-.,$&()/0123456789";

   for(j = 0; j < strlen(characters); j++, i++){
      toInt_.insert(make_pair(characters[j], i));
      toChar_.insert(make_pair(i,characters[j])); 
   }

   /* Also, map the capital letters*/

   temp = (int)'A';
   for(; temp<=(int)'Z'; i++, temp++){
      toInt_.insert(make_pair((char)temp, i));
      toChar_.insert(make_pair(i,(char)temp)); 
   }

   /* total number of the characters in our alphabet! */
   
   N_ = i;     
}

/**************************************************************************/

INT RSA::convertToInt_(string str)
{
   const char * text = str.c_str();
   int temp, length;
   double numericEqv = 0;

   length = strlen(text);

   //Fill in the loop to convert a number to an integer here

   return (INT)numericEqv;
}

/**************************************************************************/

string RSA::convertToStr_(INT num, int charNum)
{
   int i, length = charNum;
   INT temp, mod;
   char * text;

   text = new char[length+1];

   //Fill in the loop to convert an integer to a number here

   return text;
}

/**************************************************************************/

pair<INT, INT> RSA::factorize_(INT n)
{
   pair<INT, INT> factors;

   INT t, pA(0), qA(0); 
   double s;
   const double eps = 1e-6;

   for( t = (INT)sqrt((double)n)+1;;t++){
      s = sqrt(pow((double)(t),2) - (double)n); 

      //Check whether you got a perfect square!      
      if(s - floor(s) < eps){
	 pA = t + (INT)s;
	 qA = t - (INT)s;
	 break;
      }
   }
   
   factors.first = pA;
   factors.second = qA;

   return factors;
}

/**************************************************************************/

INT RSA::getInverse_(INT e, INT phi)
{

   // Fill in method for finding the multiplicative inverse of e mod phi

}

/**************************************************************************/

INT RSA::getPower_(INT val, INT power, INT mod)
{

   // Calaculate val^power mod "mod"

}

/**************************************************************************/

pair<INT,INT> RSA::generatePrimeNumbers_()
{
   /* this function will create a pair of primes(say p&q) with the property
      N^k<p*q<N^l. (N:size of the alphabet, k: encryption block size
      l: decyrption block size also assuming that both sqrt(N^k)<p&q<sqrt(N^l).
   */

   int i, j, k, l, bitNum1(0), bitNum2(0);
   INT p[2];
   INT temp, temp1, temp2;

   k = encBlockSize_;
   l = decBlockSize_;

   temp = (INT)pow((double)N_, k);

   while(temp>pow(2.0, bitNum1)){
      bitNum1++;
   }

   temp = (INT)pow((double)N_,l);
   while(temp>pow(2.0, bitNum2)){
      bitNum2++;
   }

   /* Get a random number with appropriate bit size */

   temp1 = (INT)pow(2.0, (bitNum1-1)/2);
   temp2 = (INT)pow(2.0, (bitNum2-1)/2);
   
   while(true){
      srand(time(0));
      for(i = 0; i < 2; i++){
	 p[i] = rand()%(temp2 - temp1 + 1) + temp1;
	 /* test for primality-use trial division method- if it is not prime
	    increase it by one till you get a prime number. */
	 while(true){	    
	    temp = (INT)sqrt((double)p[i]);
	    for(j = 2; j<=temp; j++){
	       if((p[i] % j) == 0)
		  break;
	    }
	    if(j>temp)
	       break;
	    p[i]++;
	 }
      }	 
      if(p[0]*p[1]>pow((double)N_, k))
	 break;
   }

   pair<INT, INT> primes;
   primes.first = p[0];
   primes.second = p[1];

   return primes;
}

/**************************************************************************/

INT RSA::getRandNumPrimeToPhi_(pair<INT, INT> primeFactors)
{
   int bitNum = 0;
   INT temp;
   INT pA = primeFactors.first, qA = primeFactors.second, e, phi;

   phi = (pA - 1) * (qA - 1);

   /* Get a prime number which will have same bit-size with phi_n */

   while(phi > pow(2.0, bitNum)){
      bitNum++;
   }

   temp = (INT)pow(2.0, bitNum - 1);

   e = (INT)(rand()%(temp - 1));
      
   for(e = e%2? e: e+1 ; e < phi; e = e + 2){
      if(gcd_(e, pA-1) == 1 && gcd_(e, qA-1) == 1){
	 break;
      }      
   }
   
   return e;   
}

/**************************************************************************/

INT RSA::gcd_(INT a, INT b)
{

   // Fill in Euclid's algorithm here
   
}

/**************************************************************************/
